#include "CAPIServer.h"

// Qt includes
#include <QTcpSocket>

CAPIServer::CAPIServer()
    : m_jetftp()
{
    connect(this,SIGNAL(newConnection()),SLOT(OnConnection()));

    // Listen on port 1025
    listen(QHostAddress::Any, 1025);

    qDebug("Listening for HTTP requests on port 1025...");
}

CAPIServer::~CAPIServer()
{
    //...
}

void CAPIServer::OnConnection()
{
    // Get the next pending connection
    QTcpSocket * pSocket = nextPendingConnection();

    // Data we read from the socket
    QString data;  // the complete contents
    QByteArray chunk;

    do
    {
        pSocket->waitForReadyRead();
        chunk = pSocket->readAll();
        data.append(chunk);

        // Also break if we read the 2x CRLF
        // that designate the end of the header.
        if(data.indexOf("\r\n\r\n") != -1)
            break;

    } while(chunk.size());

    // Now break down the HTTP request into its principle
    // components.

    // Get the first line of the HTTP header
    int iFirstSpace, iSecondSpace;

    if((iFirstSpace = data.indexOf(" ")) != -1)
        iSecondSpace = data.indexOf(" ", iFirstSpace + 1);

    if((iFirstSpace == -1) || (iSecondSpace == -1))
    {
        delete pSocket;
        return;
    }

    QString resource = data.mid(iFirstSpace + 1,(iSecondSpace - iFirstSpace - 1));

    qDebug(tr("API request: %1").arg(resource).toUtf8().data());

    QString result_str = ProcessRequest(resource);
    QByteArray raw_result = result_str.toUtf8();

    // We need to write a VERY basic header
    // to wrap the result we need to send
    int iContentLength = raw_result.size();

    pSocket->write("HTTP/1.1 200 OK\r\n");
    pSocket->write("Content-Type: text/html\r\n");
    pSocket->write(tr("Content-Length: %1\r\n\r\n").arg(iContentLength).toUtf8().data());
    pSocket->write(raw_result);

    pSocket->waitForBytesWritten();
    pSocket->close();

    delete pSocket;
}

QString CAPIServer::JSONError(QString error)
{
    return tr("{\"error\":\"%1\"}").arg(error);
}

QString CAPIServer::XMLError(QString error)
{
    return tr("<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n<error>%1</error>").arg(error);
}

QString CAPIServer::Error(QString error, bool bJSON)
{
    return (bJSON)?JSONError(error):XMLError(error);
}

QString CAPIServer::Success(bool bJSON)
{
    if(bJSON)
        return "{\"status\":\"The operation was performed successfully.\"}";
    else
        return "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n<status>The operation was performed successfully.</status>";
}

QString CAPIServer::ProcessRequest(QString request)
{
    bool bJSONP = false;
    QString jsonp_start, jsonp_end;

    // First check to see if this is a JSONP request
    if(request.indexOf("?") != -1)
    {
        // We have a JSONP request, so find the index
        // of the question mark and chop it off at that
        // point.
        int question_index = request.indexOf("?");
        QString query_string = request.right(request.size() - question_index - 1);

        // Trim the ? off the main string
        request.resize(question_index);

        QStringList params = query_string.split("=");
        if(params.size() < 2)
            return "Invalid query string.";

        // Now make sure it's the query string parameter
        if(params[0] != "jsonp")
            return "Invalid query string parameter.";

        bJSONP = true;
        jsonp_start = params[1] + "(";
        jsonp_end   = ")";
    }

    // Split the request into tokens
    QStringList tokens = request.split('/',QString::SkipEmptyParts);
    if(tokens.size() < 2)
        return "Invalid API request.";

    bool bJSON = false;  // whether we return JSON or XML
    if(tokens[0] == "json") bJSON = true;

    if(tokens.size() < 3)
        return Error("Invalid API request.",bJSON);

    // Now determine the component we're after.
    if(tokens[1] == "jetserver")
        return jsonp_start + ProcessJetServerRequest(tokens,bJSON) + jsonp_end;
    else if(tokens[1] == "jetftp")
        return jsonp_start + ProcessJetFTPRequest(tokens,bJSON) + jsonp_end;

    return Error("Invalid component specified.", bJSON);
}

QString CAPIServer::ProcessJetServerRequest(QStringList tokens, bool bJSON)
{
    if(tokens[2] == "sockets")
    {
        if(tokens.size() < 4)
            return Error("Invalid API request.",bJSON);

        CApplicationServer * pServer = NULL;
        if(tokens[3] == "jetftp") pServer = &m_jetftp;

        if(!pServer)
            return Error("Invalid component specified.",bJSON);

        // Now get the connections for that token.
        QList<int> connections = pServer->GetSocketNumbers();

        // Enumerate them and send that to the client
        QString result;

        result = (bJSON)?"{\"connections\":[" : "<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n<connection>";

        for(int i=0;i<connections.size();++i)
        {
            if(i)
                result += (bJSON)?",":"</connection>\n<connection>";

            result += tr("%1").arg(connections[i]);
        }

        result += (bJSON)?"]}":"</connection>";
        return result;
    }

    return Error("Invalid command specified.", bJSON);
}

QString CAPIServer::ProcessJetFTPRequest(QStringList tokens, bool bJSON)
{
    int conn_num = tokens[2].toInt();

    CApplicationSocket * pSocket = m_jetftp.GetSocket(conn_num);

    if(pSocket == NULL)
        return Error("The specified socket number is invalid.",bJSON);

    // Make sure we have a command
    if(tokens.size() < 4)
        return Error("Command missing.",bJSON);

    if(tokens[3] == "terminate")
        pSocket->SendMessage("TERMINATE");
    else if(tokens[3] == "getnumclients")
    {
        CJetFTPSocket * pFTPSocket = dynamic_cast<CJetFTPSocket *>(pSocket);
        int iNum = pFTPSocket->GetNumClients();

        return (bJSON)? tr("{\"num_connections\":%1}").arg(iNum):
                        tr("<?xml version=\"1.0\" encoding=\"UTF-8\" ?><num_connections>%1</num_connections>").arg(iNum);
    }

    // Then we return the results...
    return Success(bJSON);
}
